/*
 * dhcp_cache.c
 *
 *  Created on: 09.01.2011
 *      Author: Roman Chebotarev
 *         Web: www.netpatch.ru
 */

#include "common_includes.h"
#include "db2dhcp_types.h"
#include "log.h"
#include "misc_functions.h"
#include "dhcp_cache.h"
#include "dhcp_process.h"

#include <search.h>

static void * cache_root;
static pthread_rwlock_t cache_lock;
static time_t cache_node_ttl;
static time_t cache_now;
static time_t cache_last_flush;

const time_t CACHE_FLUSH_PERIOD = 60;

inline static void cache_rdlock(void);
inline static void cache_wrlock(void);
inline static void cache_unlock(void);

typedef struct dhcp_cache_flushing_queue
{
	dhcp_cache_node_t * node;
	struct dhcp_cache_flushing_queue * next;
} dhcp_fqueue_t;

static dhcp_fqueue_t * deleting_queue;

int dhcp_cache_init(time_t nodes_ttl)
{
    int err;
    if( (err = pthread_rwlock_init(&cache_lock, NULL) ) )
    {
        log_wr(CLOG, "Can't init rw_lock for DHCP cache: '%s'", strerror(err));
        return 0;
    }

    cache_node_ttl = nodes_ttl;
    cache_last_flush = time(NULL);

    return 1;
}

static int cmp_nodes(const void * n1, const void * n2)
{
	dhcp_cache_node_t *r1 = (dhcp_cache_node_t*) n1;
	dhcp_cache_node_t *r2 = (dhcp_cache_node_t*) n2;

	int ret = memcmp(r1->cli_ethaddr, r2->cli_ethaddr, ETHER_ALEN);
	if(ret)
		return ret;

	return memcmp(&r1->if_ipaddr, &r2->if_ipaddr, sizeof(r1->if_ipaddr));
}

static void action_del(const void *nodep, const VISIT which, const int depth)
{
	dhcp_cache_node_t * node = *(dhcp_cache_node_t **) nodep;

	switch (which)
	{
		case preorder:
		break;
		case postorder:
		break;
		case endorder:
		case leaf:
			if(node->timestamp + cache_node_ttl < cache_now)
			{
				char str_ether[STR_ETHER_ALEN + 1];
				char str_ipaddr[IP4_MAXSTR_ALEN + 1];

				etheraddr_bin_to_str(node->cli_ethaddr, str_ether);
				iptos(node->cached_response.dhcp_data.you_iaddr.s_addr, str_ipaddr);

				log_wr(DLOG, "Adding DHCP cache node %s/%s to deleting queue - TTL exceeded.",
						str_ether, str_ipaddr);

				dhcp_fqueue_t * del_node = calloc(1, sizeof(dhcp_fqueue_t));
				if(!del_node)
				{
					log_wr(CLOG, "Can't allocate memory for adding node to deleting queue: '%s'", strerror(errno));
					exit(error_memory);
				}
				
				del_node->node = node;

				if(deleting_queue)
				{
					del_node->next = deleting_queue;
					deleting_queue = del_node;
				}
				else	/* Empty queue */
					deleting_queue = del_node;
			}
		break;
	}
}


void dhcp_cache_flush_old(void)
{
	cache_wrlock();

	cache_now = time(NULL);

	if(cache_last_flush + CACHE_FLUSH_PERIOD > cache_now)
	{
		cache_unlock();
		return;
	}

	log_wr(DLOG, "Flushing cache: last flush ts - %lu, flush period - %lu, now is %lu.",
			cache_last_flush, CACHE_FLUSH_PERIOD, cache_now);

	deleting_queue = NULL;	/* Set start pointer to NULL */

	/* Search obsolete nodes */
	twalk(cache_root, action_del);

	/* Removing him if exists */
	dhcp_fqueue_t * q_ptr = deleting_queue;

	size_t num_del = 0;
	char str_ether[STR_ETHER_ALEN + 1];
	char str_ipaddr[IP4_MAXSTR_ALEN + 1];

	while(deleting_queue)
	{
		if(!tdelete(deleting_queue->node, &cache_root, cmp_nodes))
		{
			etheraddr_bin_to_str(deleting_queue->node->cli_ethaddr, str_ether);
			iptos(deleting_queue->node->cached_response.dhcp_data.you_iaddr.s_addr, str_ipaddr);
			log_wr(ELOG, "Can't delete node from cache: %s/%s", str_ether, str_ipaddr);
			goto _next_node;
		}
		else
			free(deleting_queue->node);
		++num_del;

_next_node:
		q_ptr = deleting_queue->next;
		free(deleting_queue);
		deleting_queue = q_ptr;
	}
  	
	log_wr(DLOG, "Cache flushed. Total %u nodes deleted.", num_del);

	cache_last_flush = cache_now;

	cache_unlock();

	return;
}


int dhcp_cache_update(const dhcp_parsed_message_t * request, const dhcp_full_packet_t * response,
		uint16_t dhcp_data_len)
{
	dhcp_cache_node_t s_node;
	s_node.cli_ethaddr = (typeof(s_node.cli_ethaddr))request->from_ether;
	s_node.if_ipaddr = request->dhcp_dev->ipaddr;

	char str_ether[STR_ETHER_ALEN + 1];
	char str_ipaddr[IP4_MAXSTR_ALEN + 1];

	etheraddr_bin_to_str(request->from_ether, str_ether);
	iptos(response->dhcp_data.you_iaddr.s_addr, str_ipaddr);

	/* Lock cache for write */
	cache_wrlock();
	time_t now = time(NULL);

	dhcp_cache_node_t * cached_node;
	dhcp_cache_node_t ** cached_node_ptr = tfind(&s_node, &cache_root, cmp_nodes);

	if(cached_node_ptr)
	{	/* Node already cached. Update. */
		cached_node = *cached_node_ptr;
		log_wr(DLOG, "Update cached data for client %s/%s.", str_ether, str_ipaddr);
		cached_node->timestamp = now;
		memcpy(&cached_node->cached_response, response, sizeof(*response));
	}
	else
	{	/* Node not found in cache. Add. */
		cached_node = calloc(1, sizeof(dhcp_cache_node_t));
		if(!cached_node)
		{
			log_wr(CLOG, "Can't allocate memory for new DHCP cache node: '%s'", strerror(errno));
			exit(error_memory);
		}

		memcpy(&cached_node->cached_response, response, sizeof(cached_node->cached_response));
		cached_node->cli_ethaddr = cached_node->cached_response.eth_head.ether_dhost;
		cached_node->if_ipaddr = request->dhcp_dev->ipaddr;
		cached_node->timestamp = now;

		if(!tsearch(cached_node, &cache_root, cmp_nodes))
		{
			log_wr(CLOG, "Can't add client %s/%s in DHCP cache. Insufficient memory?",
					str_ether, str_ipaddr);
			exit(error_memory);
		}
		log_wr(DLOG, "Added response for client %s/%s to DHCP cache.", str_ether, str_ipaddr);
	}

	/* Set DHCPACK message type for cached response */
	uint16_t type_len;
	uint8_t * cached_response_type = get_dhcp_option_ptr(&cached_node->cached_response.dhcp_data,
			cached_node->cached_response.udp_header.len, DHCP_OPT_MESSAGE_TYPE, &type_len);
	if(!cached_response_type)
	{
		log_wr(CLOG, "Invalid DHCP message cached (%s/%s): DHCP message type option not found.",
				str_ether, str_ipaddr);
		tdelete(cached_node, &cache_root, cmp_nodes);
		free(cached_node);

		cache_unlock();

		return 0;
	}

	cached_node->dhcp_data_len = dhcp_data_len;

	*cached_response_type = DHCPACK;

	cache_unlock();

	return 1;
}

dhcp_full_packet_t * dhcp_cache_find(const dhcp_parsed_message_t * request,
		dhcp_full_packet_t * response, size_t * dhcp_data_len)
{
	dhcp_cache_node_t s_node;
	s_node.cli_ethaddr = (typeof(s_node.cli_ethaddr))request->from_ether;
	s_node.if_ipaddr = request->dhcp_dev->ipaddr;

	/* Lock cache for read */
	cache_rdlock();

	dhcp_cache_node_t ** cached_node_ptr = tfind(&s_node, &cache_root, cmp_nodes);

	if(cached_node_ptr)
	{
		*dhcp_data_len = (*cached_node_ptr)->dhcp_data_len;
		memcpy(response, &(*cached_node_ptr)->cached_response, *dhcp_data_len);
	}
	else
	{
		response = NULL;
		*dhcp_data_len = 0;
	}

	cache_unlock();
	return response;
}

inline static void cache_rdlock(void)
{
    int err;
    if( (err = pthread_rwlock_rdlock(&cache_lock)) )
    {
        log_wr(CLOG, "Error on pthread_rdlock_wrlock() for lock DHCP cache: '%s'",
        		strerror(err));
        exit(error_abnormal);
    }
}

inline static void cache_wrlock(void)
{
    int err;
    if( (err = pthread_rwlock_wrlock(&cache_lock)) )
    {
        log_wr(CLOG, "Error on pthread_rwlock_wrlock() for lock DHCP cache: '%s'",
        		strerror(err));
        exit(error_abnormal);
    }
}

inline static void cache_unlock(void)
{
    int err;
    if( (err = pthread_rwlock_unlock(&cache_lock)) )
    {
        log_wr(CLOG, "Error on pthread_rwlock_unlock() for unlock DHCP cache: '%s'",
        		strerror(err));
        exit(error_abnormal);
    }
}
